from __future__ import print_function, division
import numpy as np
import os
import matplotlib.pyplot as plt
from ultranest.utils import create_logger
from ultranest import ReactiveNestedSampler
from ultranest.mlfriends import MLFriends

here = os.path.dirname(__file__)


def test_clustering():
    from ultranest.mlfriends import update_clusters
    for i in range(5):
        np.random.seed(i * 100)

        points = np.random.uniform(size=(100,2))

        nclusters, clusteridxs, overlapped_points = update_clusters(points, points, 0.1**2)

        for i in np.unique(clusteridxs):
            x, y = points[clusteridxs == i].transpose()
            plt.scatter(x, y)
        plt.savefig('testclustering_0p1.pdf', bbox_inches='tight')
        plt.close()
        assert 1 < nclusters < 30

        nclusters, clusteridxs, overlapped_points = update_clusters(points, points, 0.2**2)

        for i in np.unique(clusteridxs):
            x, y = points[clusteridxs == i].transpose()
            plt.scatter(x, y)
        plt.savefig('testclustering_0p2.pdf', bbox_inches='tight')
        plt.close()
        assert 1 <= nclusters < 2


def test_clusteringcase():
    from ultranest.mlfriends import update_clusters
    here = os.path.dirname(__file__)
    points = np.loadtxt(os.path.join(here, "clusters2.txt"))
    maxr = np.loadtxt(os.path.join(here, "clusters2_radius.txt"))
    # transformLayer = ScalingLayer()
    # transformLayer.optimize(points)
    # region = MLFriends(points, transformLayer)
    # maxr = region.compute_maxradiussq(nbootstraps=30)
    print('maxradius:', maxr)
    nclusters, clusteridxs, overlapped_points = update_clusters(points, points, maxr)

    plt.title('nclusters: %d' % nclusters)
    for i in np.unique(clusteridxs):
        x, y = points[clusteridxs == i].transpose()
        plt.scatter(x, y)
    plt.savefig('testclustering_2.pdf', bbox_inches='tight')
    plt.close()


def test_clusteringcase_eggbox():
    from ultranest.mlfriends import update_clusters, ScalingLayer, MLFriends
    points = np.loadtxt(os.path.join(here, "eggboxregion.txt"))
    transformLayer = ScalingLayer()
    transformLayer.optimize(points, points)
    region = MLFriends(points, transformLayer)
    maxr = region.compute_maxradiussq(nbootstraps=30)
    assert 1e-10 < maxr < 5e-10
    print('maxradius:', maxr)
    nclusters, clusteridxs, overlapped_points = update_clusters(points, points, maxr)
    # plt.title('nclusters: %d' % nclusters)
    # for i in np.unique(clusteridxs):
    #    x, y = points[clusteridxs == i].transpose()
    #    plt.scatter(x, y)
    # plt.savefig('testclustering_eggbox.pdf', bbox_inches='tight')
    # plt.close()
    assert 14 < nclusters < 20, nclusters


class MockIntegrator(ReactiveNestedSampler):
    def __init__(self):
        self.use_mpi = False
        self.mpi_size = 1
        self.mpi_rank = 0
        self.region = None
        self.transformLayer = None
        self.wrapped_axes = []
        self.log = True
        self.logger = create_logger("mock")
        self.region_class = MLFriends


def test_overclustering_eggbox_txt():
    from ultranest.mlfriends import update_clusters, ScalingLayer, MLFriends
    np.random.seed(1)
    for i in [20, 23, 24, 27, 49]:
        print()
        print("==== TEST CASE %d =====================" % i)
        print()
        points = np.loadtxt(os.path.join(here, "overclustered_u_%d.txt" % i))

        for k in range(3):
            transformLayer = ScalingLayer(wrapped_dims=[])
            transformLayer.optimize(points, points)
            region = MLFriends(points, transformLayer)
            maxr = region.compute_maxradiussq(nbootstraps=30)
            region.maxradiussq = maxr
            nclusters = transformLayer.nclusters

            print("manual: r=%e nc=%d" % (region.maxradiussq, nclusters))
            # assert 1e-10 < maxr < 5e-10
            nclusters, clusteridxs, overlapped_points = update_clusters(points, points, maxr)
            print("reclustered: nc=%d" % (nclusters))

        if False:
            plt.title('nclusters: %d' % nclusters)
            for k in np.unique(clusteridxs):
                x, y = points[clusteridxs == k].transpose()
                plt.scatter(x, y)
            plt.savefig('testoverclustering_eggbox_%d.pdf' % i, bbox_inches='tight')
            plt.close()
        assert 14 < nclusters < 20, (nclusters, i)

        for j in range(3):
            nclusters, clusteridxs, overlapped_points = update_clusters(points, points, maxr)
            assert 14 < nclusters < 20, (nclusters, i)


def test_overclustering_eggbox_update(plot=False):
    np.random.seed(1)
    for i in [20, 23, 24, 27, 42]:
        print()
        print("==== TEST CASE %d =====================" % i)
        print()
        mock = MockIntegrator()
        print("loading...")
        data = np.load(os.path.join(here, "overclustered_%d.npz" % i))
        print("loading... done")

        nsamples, mock.x_dim = data['u0'].shape
        noverlap = 0
        for i, u1 in enumerate(data['u']):
            assert len((u1 == data['u0']).all(axis=1)) == nsamples
            noverlap += (u1 == data['u0']).all(axis=1).sum()
        print("u0:%d -> u:%d : %d points are common" % (nsamples, nsamples, noverlap))

        mock._update_region(data['u0'], data['u0'])
        nclusters = mock.transformLayer.nclusters
        print("initialised with: r=%e nc=%d" % (mock.region.maxradiussq, nclusters))
        smallest_cluster = min(
            (mock.transformLayer.clusterids == i).sum()
            for i in np.unique(mock.transformLayer.clusterids))
        if smallest_cluster == 1:
            print("found lonely points")

        print(" --- intermediate tests how create_new reacts ---")
        nextTransformLayer = mock.transformLayer.create_new(data['u0'], mock.region.maxradiussq)
        print("updated to (with same data): r=%e nc=%d" % (mock.region.maxradiussq, nclusters))
        smallest_cluster = min((nextTransformLayer.clusterids == i).sum() for i in np.unique(nextTransformLayer.clusterids))
        assert smallest_cluster > 1, ("found lonely points", i, nclusters, np.unique(mock.transformLayer.clusterids, return_counts=True))

        nextTransformLayer = mock.transformLayer.create_new(data['u'], mock.region.maxradiussq)
        nclusters = nextTransformLayer.nclusters
        print("updated to (with new data): r=%e nc=%d" % (mock.region.maxradiussq, nclusters))
        smallest_cluster = min((nextTransformLayer.clusterids == i).sum() for i in np.unique(nextTransformLayer.clusterids))
        if smallest_cluster > 1:
            # this happens because mock.region.maxradiussq is not valid anymore
            # when nlive changes
            print("found lonely points", i, nclusters, np.unique(mock.transformLayer.clusterids, return_counts=True))

        if plot:
            for xi0, yi0, xi, yi in zip(data['u0'][:,0], data['u0'][:,1], data['u'][:,0], data['u'][:,1]):
                plt.plot([xi0, xi], [yi0, yi], 'x-', ms=2)

            plt.savefig('testoverclustering_eggbox_%d_diff.pdf' % i, bbox_inches='tight')
            plt.close()
        print(" --- end ---")

        if len(data['u']) < nsamples or True:
            # maxradius has to be invalidated if live points change
            print("setting maxradiussq to None")
            mock.region.maxradiussq = None

        updated = mock._update_region(data['u'], data['u'])
        nclusters = mock.transformLayer.nclusters
        print("transitioned to : r=%e nc=%d %s" % (mock.region.maxradiussq, nclusters, updated))
        smallest_cluster = min((mock.transformLayer.clusterids == i).sum() for i in np.unique(mock.transformLayer.clusterids))
        if smallest_cluster == 1:
            print("found lonely points")
        for k in np.unique(mock.transformLayer.clusterids):
            x, y = mock.region.u[mock.transformLayer.clusterids == k].transpose()
            print('cluster %d/%d: %d points @ %.5f +- %.5f , %.5f +- %.5f' % (k, nclusters, len(x), x.mean(), x.std(), y.mean(), y.std()))
        if plot:
            plt.title('nclusters: %d' % nclusters)
            for k in np.unique(mock.transformLayer.clusterids):
                x, y = mock.region.u[mock.transformLayer.clusterids == k].transpose()
                plt.scatter(x, y, s=2)

            plt.savefig('testoverclustering_eggbox_%d.pdf' % i, bbox_inches='tight')
            plt.close()
        assert 14 < nclusters < 20, (nclusters, i)
        assert smallest_cluster > 1, (i, nclusters, np.unique(mock.transformLayer.clusterids, return_counts=True))


if __name__ == '__main__':
    # test_clustering()
    # test_clusteringcase()
    test_overclustering_eggbox_update(plot=True)



from hypothesis import given
import hypothesis as h
import hypothesis.strategies as st
import numpy as np
import matplotlib.pyplot as plt
import time
import ultranest.mlfriends
from ultranest.mlfriends import pareto_front_filter, find_slope

MEPS = 1 + 1e-10
AEPS = 1e-10

def find_bounding_line_exhaustive(points):
    assert np.isfinite(points).all(), points
    indices = np.argsort(points[:,0])
    x = points[indices,0]
    y = points[indices,1]
    del points
    xmin = x[0]
    xmax = x[-1]

    yminopt = y[0]
    ymaxopt = y[0]
    Aopt = np.inf
    for j in range(len(x)):
        xj, yj = x[j], y[j]
        for i in range(j):
            assert i < j, (i, j)
            # for xi, yi in zip(x[:k], y[:k]):
            xi, yi = x[i], y[i]
            # i is left of j
            assert xi <= xj
            if yi > yj:
                # slope would be negative
                continue
            slope = (yj - yi) / (xj - xi)
            assert slope >= 0, slope
            # make interpolation 
            enclosed = y <= ((x - xi) * slope + yi) * MEPS
            # enclosed[j] = True
            if not enclosed.all():
                #print('1:', A, 'not enclosing all:', xi, yi, xj, yj, ':', x[~enclosed][0], y[~enclosed][0])
                continue
            
            ymin = (xmin - xi) * slope + yi
            ymax = (xmax - xi) * slope + yi
            A = (ymax + ymin) / 2
            #print('1:', A)
            # check if all points are contained
            if A >= 0 and A < Aopt:
                yminopt, ymaxopt, Aopt = ymin, ymax, A
            #elif A < Aopt:
            #   yminopt, ymaxopt = ymin, ymin
    #print('1: Aopt', Aopt)
    assert np.isfinite(yminopt)
    assert np.isfinite(ymaxopt)
    return yminopt, ymaxopt, xmin, xmax

def pareto_front(points):
    # pareto filtering:
    at_front = np.ones(len(points), dtype=bool)
    for xi, yi in points:
        dominated = np.logical_and(points[:,0] > xi, points[:,1] <= yi)
        at_front = np.logical_and(at_front, ~dominated)
    return at_front

def pareto_front_sorted(points):
    # pareto filtering:
    at_front = np.ones(len(points), dtype=bool)
    for i, (xi, yi) in enumerate(points):
        dominated = np.logical_and(points[i+1:,0] > xi, points[i+1:,1] <= yi)
        at_front[i+1:] = np.logical_and(at_front[i+1:], ~dominated)
        #dominated = np.logical_and(points[:,0] > xi, points[:,1] <= yi)
        #at_front = np.logical_and(at_front, ~dominated)
    return at_front

def pareto_front_sorted2(points):
    # pareto filtering:
    at_front = np.ones(len(points), dtype=bool)
    y = points[:, 1]
    #for i, yi in enumerate(y[:-1]):
    for i in range(len(y)-1):
        if not at_front[i]: continue
        #at_front[i+1:] = np.logical_and(at_front[i+1:], y[i+1:] > y[i])
        at_front[i+1:][y[i+1:] <= y[i]] = False
        #dominated = np.logical_and(points[:,0] > xi, points[:,1] <= yi)
        #at_front = np.logical_and(at_front, ~dominated)
    return at_front

def find_bounding_line_exhaustive_pareto(points, sorted=False):
    xmin = points[:,0].min()
    xmax = points[:,0].max()
    assert np.isfinite(points).all()
    points = points[pareto_front(points),:]
    if len(points) == 1:
        return points[0,1], points[0,1], xmin, xmax
    assert np.isfinite(points).all()

    if sorted:
        x = points[:,0]
        y = points[:,1]
    else:
        indices = np.argsort(points[:,0])
        x = points[indices,0]
        y = points[indices,1]
    del points
    yminopt = np.nan
    ymaxopt = np.nan
    Aopt = np.inf
    for j in range(len(x)):
        for i in range(j):
            # i is left of j
            assert x[i] <= x[j]
            # make interpolation 
            slope = (y[j] - y[i]) / (x[j] - x[i])
            ypred = (x - x[i]) * slope + y[i]
            enclosed = y <= ypred * MEPS
            #print(i, j, enclosed, y, (x - x[i]), (y[j] - y[i]) / (x[j] - x[i]), y[i])
            assert enclosed[i], (y[i], ypred[i], ypred[i] * MEPS, x, y)
            enclosed[j] = True
            #assert enclosed[j], (y[j], (x[j] - x[i]) * (y[j] - y[i]) / (x[j] - x[i]) + y[i])
            if not enclosed.all():
                continue
            
            ymin = (xmin - x[i]) * slope + y[i]
            ymax = (xmax - x[i]) * slope + y[i]
            A = (ymax + ymin) / 2
            #print('2:', A, x[i], y[i], x[j], y[j])
            # check if all points are contained
            if A >= 0 and A < Aopt:
                yminopt, ymaxopt, Aopt = ymin, ymax, A

    #print('2: Aopt:', Aopt)
    assert np.isfinite(yminopt), (yminopt, x, y)
    assert np.isfinite(ymaxopt), (ymaxopt, x, y)
    return yminopt, ymaxopt, xmin, xmax


def find_bounding_line_exhaustive_pareto_vectorized1(points, sorted=False):
    if sorted:
        sorted_points = points
    else:
        indices = np.argsort(points[:,0])
        sorted_points = points[indices,:]
    del points
    xmin, xmax = sorted_points[0,0], sorted_points[-1,0]
    
    points = sorted_points[pareto_front_sorted(sorted_points),:]
    del sorted_points
    if len(points) == 1:
        return points[0,1], points[0,1], xmin, xmax

    x = points[:,0]
    y = points[:,1]
    
    yminopt = np.nan
    ymaxopt = np.nan
    Aopt = np.inf
    for k, (xi, yi) in enumerate(points[:-1,:]):
        # i is left of j
        xj, yj = x[k+1:].reshape((-1, 1)), y[k+1:].reshape((-1, 1))
        slopes = (yj - yi) / (xj - xi)
        
        # make interpolation
        mask_enclosed = (y.reshape((1, -1)) <= MEPS * ((x.reshape((1, -1)) - xi) * slopes + yi)).all(axis=1)
        assert mask_enclosed.shape == (len(slopes),), (mask_enclosed.shape, slopes.shape)
        mask_enclosed = np.logical_and(mask_enclosed, slopes[:,0] >= 0)
        assert mask_enclosed.shape == (len(slopes),), (mask_enclosed.shape, slopes.shape)
        if not mask_enclosed.any():
            continue
        
        # xi, yi, slopes = xi[mask_enclosed,0], yi[mask_enclosed,0], slopes[mask_enclosed,0]
        slopes = slopes[mask_enclosed,0]
        
        ymin = (xmin - xi) * slopes + yi
        ymax = (xmax - xi) * slopes + yi
        A = (ymax + ymin) / 2
        # check if all points are contained
        if not (A < Aopt).any():
            continue
        i = np.argmin(A)
        assert A[i] >= 0 and A[i] < Aopt, (A[i], Aopt)
        yminopt, ymaxopt, Aopt = ymin[i], ymax[i], A[i]

    return yminopt, ymaxopt, xmin, xmax

def find_bounding_line_exhaustive_pareto_vectorized2(points, sorted=False):
    if sorted:
        sorted_points = points
    else:
        indices = np.argsort(points[:,0])
        sorted_points = points[indices,:]
    del points
    xmin, xmax = sorted_points[0,0], sorted_points[-1,0]
    
    points = sorted_points[pareto_front_sorted(sorted_points),:]
    del sorted_points
    if len(points) == 1:
        return points[0,1], points[0,1], xmin, xmax

    x = points[:,0]
    y = points[:,1]
    
    yminopt = np.nan
    ymaxopt = np.nan
    Aopt = np.inf
    for j in range(len(x)):
        for i in range(j):
            # i is left of j
            assert x[i] <= x[j]
            # make interpolation
            slope = (y[j] - y[i]) / (x[j] - x[i])
            enclosed = y <= ((x - x[i]) * slope + y[i]) * MEPS
            if not enclosed.all():
                continue
            
            ymin = (xmin - x[i]) * slope + y[i]
            ymax = (xmax - x[i]) * slope + y[i]
            A = (ymax + ymin) / 2
            # check if all points are contained
            if A >= 0 and A < Aopt:
                yminopt, ymaxopt, Aopt = ymin, ymax, A

    return yminopt, ymaxopt, xmin, xmax


find_bounding_line_exhaustive_pareto_vectorized = find_bounding_line_exhaustive_pareto_vectorized2

def orig_test_pareto(N, k, r):
    points = np.transpose([
        np.random.uniform(0, 1, size=N),
        k * np.random.uniform(0, 1, size=N) + 0.01,
    ])
    points.flags.writeable = False
    ppoints = points[pareto_front(points),:]
    sorted_ppoints = ppoints[np.argsort(ppoints[:,0]),:]
    sorted_points = points[np.argsort(points[:,0]),:]
    pspoints = sorted_points[pareto_front_sorted(sorted_points),:]
    pspoints2 = sorted_points[pareto_front_sorted2(sorted_points),:]
    pspoints3 = sorted_points[pareto_front_filter(sorted_points[:,0], sorted_points[:,1]),:]
    np.testing.assert_allclose(sorted_ppoints, pspoints)
    np.testing.assert_allclose(sorted_ppoints, pspoints2)
    np.testing.assert_allclose(sorted_ppoints, pspoints3)

#@h.settings(max_examples=5000)
@given(st.integers(min_value=2, max_value=1000), st.floats(min_value=0.0, max_value=1e5), st.random_module())
def test_pareto(N, k, r):
    return orig_test_pareto(N, k, r)

def speedtest_pareto():
    slopes = np.linspace(0, 1, 20)
    for N in [10, 40, 100, 400, 1000]:
        inputs = []
        for k in slopes:
            points = np.transpose([
                np.random.uniform(0, 1, size=N),
                k * np.random.uniform(0, 1, size=N) + 0.01,
            ])
            sorted_points = points[np.argsort(points[:,0]),:]
            inputs.append(sorted_points)

        t0 = time.time()
        for points in inputs:
            for i in range(10000 // N):
                pareto_front(points)
        t1 = time.time() - t0
        t0 = time.time()
        for points in inputs:
            for i in range(10000 // N):
                pareto_front_sorted(sorted_points)
        t2 = time.time() - t0
        t0 = time.time()
        for points in inputs:
            for i in range(100000 // N):
                pareto_front_sorted2(sorted_points)
        t3 = time.time() - t0
        t0 = time.time()
        for points in inputs:
            for i in range(100000 // N):
                pareto_front_filter(sorted_points[:,0], sorted_points[:,1])
        t4 = time.time() - t0
        print("%d %.3fs %.3fs %.3fs %.3fs (pareto filter)" % (N, t1 * 10, t2 * 10, t3, t4))

def speedtest_bounding_line():
    slopes = np.linspace(0, 1, 20)
    for N in [10, 40, 100, 400, 1000]:
        indicator = np.zeros(N, dtype=bool)
        inputs = []
        for k in slopes:
            points = np.transpose([
                np.random.uniform(0, 1, size=N),
                k * np.random.uniform(0, 1, size=N) + 0.01,
            ])
            inputs.append(points)
        results = []
        t0 = time.time()
        for points in inputs:
            results.append(find_bounding_line_exhaustive(points))
        texhaustive = time.time() - t0
        results2 = []
        t0 = time.time()
        for points in inputs:
            for i in range(100-1):
                find_bounding_line_exhaustive_pareto(points)
            results2.append(find_bounding_line_exhaustive_pareto(points))
        tpareto = time.time() - t0
        results3 = []
        t0 = time.time()
        for points in inputs:
            for i in range(100-1):
                find_bounding_line_exhaustive_pareto_vectorized1(points)
            results3.append(find_bounding_line_exhaustive_pareto_vectorized1(points))
        tparetov = time.time() - t0
        results4 = []
        t0 = time.time()
        for points in inputs:
            for i in range(100-1):
                find_bounding_line_exhaustive_pareto_vectorized2(points)
            results4.append(find_bounding_line_exhaustive_pareto_vectorized2(points))
        tparetov2 = time.time() - t0
        results5 = []
        t0 = time.time()
        for points in inputs:
            for i in range(1000-1):
                indices = np.argsort(points[:,0])
                sorted_x = points[indices,0]
                sorted_y = points[indices,1]
                find_slope(sorted_x, sorted_y, indicator)
            results5.append(find_slope(sorted_x, sorted_y, indicator)[:-1])
        tparetoc = time.time() - t0
        print("%d %.3fs %.3fs %.3fs %.3fs %.3fs (find_slope)" % (N, texhaustive, tpareto, tparetov, tparetov2, tparetoc / 10))
        for input, k, r1, r2 in zip(inputs, slopes, results, results2):
            np.testing.assert_allclose(r1, r2, err_msg="input slope: %s" % k)
        for input, k, r1, r3 in zip(inputs, slopes, results, results3):
            np.testing.assert_allclose(r1, r3, err_msg="input slope: %s" % k)
        for input, k, r1, r4 in zip(inputs, slopes, results, results4):
            np.testing.assert_allclose(r1, r4, err_msg="input slope: %s" % k)
        for input, k, r1, r5 in zip(inputs, slopes, results, results5):
            np.testing.assert_allclose(r1, r5, err_msg="input slope: %s" % k)


def speedtest_bounding_line_multid():
    print("speedtest_bounding_line_multid")
    d = 50
    use_sorted = True
    for N in [40, 100, 400, 1000]:
        indicator = np.ones(N, dtype=bool)
        slopes = np.linspace(0, 1, 2000 // N)
        inputs = []
        chunks = []
        for k in slopes:
            points = np.transpose([np.random.uniform(0, 1, size=N)] +
                [k * np.random.uniform(0, 1, size=N) + 0.01 for i in range(d)])
            sorted_points = points[np.argsort(points[:,0]),:]
            del points
            sorted_points.flags.writeable = False
            chunk = []
            for r in range(20):
                selected = np.zeros(N, dtype=bool)
                selected[np.random.randint(0, N, size=N)] = True
                #if selected.sum() > 1:
                chunk.append(sorted_points[selected,:])
            inputs += chunk
            chunks.append(chunk)
        results = []
        t0 = time.time()
        for points in inputs:
            for i in range(d):
                results.append(find_bounding_line_exhaustive_pareto(points[:,[0,1+i]], sorted=use_sorted))
        tpareto = time.time() - t0
        results2 = []
        t0 = time.time()
        for points in inputs:
            for i in range(d):
                results2.append(find_bounding_line_exhaustive_pareto_vectorized1(points[:,[0,1+i]], sorted=use_sorted))
        tparetov = time.time() - t0
        results3 = []
        t0 = time.time()
        for points in inputs:
            for i in range(d):
                results3.append(find_bounding_line_exhaustive_pareto_vectorized2(points[:,[0,1+i]], sorted=use_sorted))
        tparetov2 = time.time() - t0
        results4 = []
        t0 = time.time()
        for _ in range(10):
            for points in inputs:
                for i in range(d):
                    results4.append(find_slope(points[:,0], points[:,i+1], indicator[:len(points)])[:-1])
        tparetoc = time.time() - t0
        print("%d %.3fs %.3fs %.3fs %.3fs (find_slope multiple%s)" % (N, tpareto, tparetov, tparetov2, tparetoc / 10, ', assuming sorted' if use_sorted else ''))
        for input, k, r1, r2 in zip(inputs, slopes, results, results2):
            np.testing.assert_allclose(r1, r2, err_msg="input slope: %s" % k)
        for input, k, r1, r3 in zip(inputs, slopes, results, results3):
            np.testing.assert_allclose(r1, r3, err_msg="input slope: %s" % k)
        for input, k, r1, r4 in zip(inputs, slopes, results, results4):
            np.testing.assert_allclose(r1, r4, err_msg="input slope: %s" % k)

def orig_test_bounding_linep(N, k, r, plot):
    points = np.transpose([
        np.random.uniform(0, 1, size=N),
        k * np.random.uniform(0, 1, size=N) + 0.01,
    ])
    points.flags.writeable = False
    #print("points:", points)
    indicator = np.ones(N, dtype=bool)
    ppoints = points[pareto_front(points),:]
    sorted_ppoints = ppoints[np.argsort(ppoints[:,0]),:]
    sorted_points = points[np.argsort(points[:,0]),:]
    pspoints = sorted_points[pareto_front_sorted(sorted_points),:]
    if plot:
        plt.plot(points[:,0], points[:,1], 'o ')
        plt.plot(ppoints[:,0], ppoints[:,1], 'x ', ms=10)
        plt.plot(pspoints[:,0], pspoints[:,1], '+ ', ms=10)
        plt.savefig('test_boundingline.pdf', bbox_inches='tight')
        plt.close()
    
    r1 = find_bounding_line_exhaustive(points)
    r2 = find_bounding_line_exhaustive_pareto(points)
    r3 = find_bounding_line_exhaustive_pareto_vectorized1(points)
    r4 = find_bounding_line_exhaustive_pareto_vectorized2(points)
    r5 = find_slope(sorted_points[:,0], sorted_points[:,1], indicator)
    ymin, ymax, xmin, xmax = r1
    ymin2, ymax2, xmin2, xmax2 = r2
    ymin3, ymax3, xmin3, xmax3 = r3
    ymin4, ymax4, xmin4, xmax4 = r4
    if plot:
        plt.plot([xmin, xmax], [ymin, ymax], '--')
        plt.plot([xmin2, xmax2], [ymin2, ymax2], alpha=0.5, lw=4)
        #plt.plot([xmin3, xmax3], [ymin3, ymax3], alpha=0.5, lw=6)
        plt.plot(points[:,0], points[:,1], 'o ')
        plt.plot(ppoints[:,0], ppoints[:,1], 'x ', ms=10)
        plt.plot(pspoints[:,0], pspoints[:,1], '+ ', ms=10)
        plt.savefig('test_boundingline.pdf', bbox_inches='tight')
        plt.close()
        time.sleep(0.2)

    np.testing.assert_allclose(sorted_ppoints, pspoints)

    ypred = (sorted_points[:,0] - xmin) / (xmax - xmin) * (ymax - ymin) + ymin
    assert (ypred * MEPS >= sorted_points[:,1]).all(), (ypred, sorted_points[:,1], ypred >= sorted_points[:,1])

    ypred = (sorted_points[:,0] - xmin2) / (xmax2 - xmin2) * (ymax2 - ymin2) + ymin2
    assert (ypred * MEPS >= sorted_points[:,1]).all(), (ypred, sorted_points[:,1], ypred >= sorted_points[:,1])

    np.testing.assert_allclose(list(r1), list(r2))

    ypred = (points[:,0] - xmin3) / (xmax3 - xmin3) * (ymax3 - ymin3) + ymin3
    assert (ypred * MEPS >= points[:,1]).all(), (ypred, points[:,1])

    np.testing.assert_allclose(list(r1), list(r3))
    np.testing.assert_allclose(list(r1), list(r4))
    np.testing.assert_allclose(list(r1), list(r5)[:-1])
    assert r5[-1] == len(pspoints)

@h.settings(deadline=10000.0, max_examples=3, verbosity=h.Verbosity.verbose)
@given(st.integers(min_value=2, max_value=1000), st.floats(min_value=0.0, max_value=1e100), st.random_module())
def test_bounding_linep(N, k, r):
    return orig_test_bounding_linep(N, k, r, plot=False)

@given(st.integers(min_value=2, max_value=100), st.floats(min_value=0.0, max_value=1), st.random_module())
def test_bounding_linep(N, k, r):
    return orig_test_bounding_linep(N, k, r, plot=False)

def test_bounding_line(plot=False):
    print("test_bounding_line")
    #np.random.seed(1)
    np.random.seed(223)
    orig_test_bounding_linep(6, 29790.523779723226, 0, plot)
    np.random.seed(0)
    orig_test_bounding_linep(256, 0.0, 0, plot)
    np.random.seed(0)
    orig_test_bounding_linep(2, 1, 0, plot)
    np.random.seed(0)
    orig_test_bounding_linep(417, 0.2293582383128268, 0, plot)
    print("test_bounding_line done")

if __name__ == '__main__':
    test_bounding_line(plot=True)
    print("testing pareto filters...")
    test_pareto()
    print("testing bounding line functions...")
    test_bounding_linep()
    print("speed tests...")
    import sys
    if 'speedtest-pareto' in sys.argv[1:]:
        speedtest_pareto()
    if 'speedtest-slopes' in sys.argv[1:]:
        speedtest_bounding_line()
    if 'speedtest-slopes-multid' in sys.argv[1:]:
        speedtest_bounding_line_multid()



import numpy as np
from ultranest.mlfriends import ScalingLayer, AffineLayer, MLFriends
from ultranest import ReactiveNestedSampler
from ultranest.stepsampler import RegionMHSampler, CubeMHSampler, CubeSliceSampler, RegionSliceSampler, SpeedVariableRegionSliceSampler, AHARMSampler, RegionBallSliceSampler
from ultranest.stepsampler import generate_region_random_direction, ellipsoid_bracket, crop_bracket_at_unit_cube
from ultranest.pathsampler import SamplingPathStepSampler
from numpy.testing import assert_allclose

#here = os.path.dirname(__file__)

def loglike_vectorized(z):
    a = np.array([-0.5 * sum([((xi - 0.7 + i*0.001)/0.1)**2 for i, xi in enumerate(x)]) for x in z])
    b = np.array([-0.5 * sum([((xi - 0.3 - i*0.001)/0.1)**2 for i, xi in enumerate(x)]) for x in z])
    return np.logaddexp(a, b)

def loglike(x):
    a = -0.5 * sum([((xi - 0.7 + i*0.001)/0.1)**2 for i, xi in enumerate(x)])
    b = -0.5 * sum([((xi - 0.3 - i*0.001)/0.1)**2 for i, xi in enumerate(x)])
    return np.logaddexp(a, b)

def transform(x):
    return x # * 10. - 5.

def transform1(x):
    return x**2

paramnames = ['param%d' % i for i in range(3)]
#paramnames = ['param%d' % i for i in range(40)]

def test_stepsampler_cubemh(plot=False):
    np.random.seed(1)
    sampler = ReactiveNestedSampler(paramnames, loglike_vectorized, transform=transform1, vectorized=True)
    sampler.stepsampler = CubeMHSampler(nsteps=4 * len(paramnames))
    r = sampler.run(log_interval=50, min_num_live_points=400)
    sampler.print_results()
    a = (np.abs(r['samples'] - 0.7) < 0.1).all(axis=1)
    b = (np.abs(r['samples'] - 0.3) < 0.1).all(axis=1)
    assert a.sum() > 1, a.sum()
    assert b.sum() > 1, b.sum()

def test_stepsampler_regionmh(plot=False):
    np.random.seed(2)
    sampler = ReactiveNestedSampler(paramnames, loglike_vectorized, transform=transform, vectorized=True)
    sampler.stepsampler = RegionMHSampler(nsteps=4 * len(paramnames))
    r = sampler.run(log_interval=50, min_num_live_points=400)
    sampler.print_results()
    a = (np.abs(r['samples'] - 0.7) < 0.1).all(axis=1)
    b = (np.abs(r['samples'] - 0.3) < 0.1).all(axis=1)
    assert a.sum() > 1, a
    assert b.sum() > 1, b

def test_stepsampler_cubeslice(plot=False):
    np.random.seed(3)
    sampler = ReactiveNestedSampler(paramnames, loglike, transform=transform1)
    sampler.stepsampler = CubeSliceSampler(nsteps=len(paramnames))
    r = sampler.run(log_interval=50, min_num_live_points=400)
    sampler.print_results()
    a = (np.abs(r['samples'] - 0.7) < 0.1).all(axis=1)
    b = (np.abs(r['samples'] - 0.3) < 0.1).all(axis=1)
    assert a.sum() > 1
    assert b.sum() > 1

def test_stepsampler_regionslice(plot=False):
    np.random.seed(4)
    sampler = ReactiveNestedSampler(paramnames, loglike, transform=transform)
    sampler.stepsampler = RegionSliceSampler(nsteps=len(paramnames))
    r = sampler.run(log_interval=50, min_num_live_points=400)
    sampler.print_results()
    a = (np.abs(r['samples'] - 0.7) < 0.1).all(axis=1)
    b = (np.abs(r['samples'] - 0.3) < 0.1).all(axis=1)
    assert a.sum() > 1
    assert b.sum() > 1


def test_stepsampler_variable_speed(plot=False):
    matrices = [
        np.array([[True, True, True], [False, True, True], [False, False, True]]),
        [Ellipsis, slice(1,None), slice(2,4)]
    ]
    for i, matrix in enumerate(matrices):
        np.random.seed(42 + i)
        sampler = ReactiveNestedSampler(paramnames, loglike, transform=transform)
        sampler.stepsampler = SpeedVariableRegionSliceSampler(matrix)
        r = sampler.run(log_interval=50, min_num_live_points=400)
        sampler.print_results()
        a = (np.abs(r['samples'] - 0.7) < 0.1).all(axis=1)
        b = (np.abs(r['samples'] - 0.3) < 0.1).all(axis=1)
        assert a.sum() > 1
        assert b.sum() > 1

def make_region(ndim, us=None):
    if us is None:
        us = np.random.uniform(size=(1000, ndim))
    
    if ndim > 1:
        transformLayer = AffineLayer()
    else:
        transformLayer = ScalingLayer()
    transformLayer.optimize(us, us)
    region = MLFriends(us, transformLayer)
    region.maxradiussq, region.enlarge = region.compute_enlargement(nbootstraps=30)
    region.create_ellipsoid(minvol=1.0)
    return region


def test_stepsampler(plot=False):
    np.random.seed(6)
    region = make_region(len(paramnames))
    Ls = loglike_vectorized(region.u)
    
    stepsampler = CubeMHSampler(nsteps=len(paramnames))
    while True:
        u1, p1, L1, nc = stepsampler.__next__(region, -1e100, region.u, Ls, transform, loglike)
        if u1 is not None:
            break
    assert L1 > -1e100
    print(u1, L1)
    while True:
        u2, p2, L2, nc = stepsampler.__next__(region, -1e100, region.u, Ls, transform, loglike)
        if u2 is not None:
            break
    assert L2 > -1e100
    print(u2, L2)
    assert np.all(u1 != u2)
    assert np.all(L1 != L2)

def test_stepsampler_adapt_when_stuck(plot=False):
    # check that a stuck sampler can free itself
    np.random.seed(7)
    us = np.random.normal(0.7, 0.0001, size=(1000, len(paramnames)))
    region = make_region(len(paramnames), us=us)
    Ls = loglike_vectorized(us)
    Lmin = Ls.min()

    print('CubeMHSampler')
    stepsampler = CubeMHSampler(nsteps=1, region_filter=True)
    np.random.seed(23)
    old_scale = stepsampler.scale
    for i in range(1000):
        if i > 100:
            assert False, i
        unew, pnew, Lnew, nc = stepsampler.__next__(region, Lmin, us, Ls, transform, loglike, ndraw=10)
        if unew is not None:
            break
    
    new_scale = stepsampler.scale
    assert new_scale != old_scale
    assert new_scale < 0.01, (new_scale, unew)
    
    print('CubeSliceSampler')
    stepsampler = CubeSliceSampler(nsteps=1, region_filter=True)
    np.random.seed(23)
    old_scale = stepsampler.scale
    for j in range(100):
        for i in range(1000):
            if i > 100:
                assert False, i
            unew, pnew, Lnew, nc = stepsampler.__next__(region, Lmin, us, Ls, transform, loglike, ndraw=10)
            if unew is not None:
                break
    
    new_scale = stepsampler.scale
    assert new_scale != old_scale
    assert new_scale < 0.01, (new_scale, unew)

def test_stepsampler_regionmh_adapt(plot=False):
    np.random.seed(8)
    region = make_region(len(paramnames))
    Ls = loglike_vectorized(region.u)
    try:
        RegionMHSampler(nsteps=len(paramnames), adaptive_nsteps='Hello')
        assert False, 'expected error'
    except ValueError:
        pass
    
    for sampler_class in RegionMHSampler, CubeMHSampler, CubeSliceSampler, RegionSliceSampler: 
        for adaptation in False, 'move-distance', 'proposal-total-distances', 'proposal-summed-distances':
            print()
            stepsampler = sampler_class(nsteps=len(paramnames), adaptive_nsteps=adaptation)
            print(stepsampler)
            stepsampler.region_changed(Ls, region)
            np.random.seed(23)
            old_scale = stepsampler.scale
            for i in range(5):
                while True:
                    unew, pnew, Lnew, nc = stepsampler.__next__(region, -1e100, region.u, Ls, transform, loglike)
                    if unew is not None:
                        break
            new_scale = stepsampler.scale
            assert new_scale != old_scale
            
            if adaptation:
                assert stepsampler.nsteps != len(paramnames)
            else:
                assert stepsampler.nsteps == len(paramnames)

def assert_point_touches_ellipsoid(ucurrent, v, t, ellipsoid_center, ellipsoid_invcov, enlarge):
    unext = ucurrent + v * t
    d = unext - ellipsoid_center
    r = np.einsum('j,jk,k->', d, ellipsoid_invcov, d)
    assert np.isclose(r, enlarge), (ucurrent, t, r, enlarge)

def test_ellipsoid_bracket(plot=False):
    for seed in range(20):
        print("seed:", seed)
        np.random.seed(seed)
        if seed % 2 == 0:
            us = np.random.normal(size=(2**np.random.randint(3, 10), 2))
            us /= ((us**2).sum(axis=1)**0.5).reshape((-1, 1))
            us = us * 0.1 + 0.5
        else:
            us = np.random.uniform(size=(2**np.random.randint(3, 10), 2))
        
        if plot:
            import matplotlib.pyplot as plt
            plt.plot(us[:,0], us[:,1], 'o ', ms=2)
        
        transformLayer = ScalingLayer()
        region = MLFriends(us, transformLayer)
        try:
            region.maxradiussq, region.enlarge = region.compute_enlargement()
            region.create_ellipsoid()
        except ValueError:
            continue

        print(region.ellipsoid_center)
        print(region.enlarge)
        print(region.ellipsoid_cov)
        print(region.ellipsoid_invcov)
        print(region.ellipsoid_axes)
        print(region.ellipsoid_inv_axes)

        ucurrent = np.array([2**0.5*0.1/2+0.5, 2**0.5*0.1/2+0.5])
        ucurrent = np.array([0.4, 0.525])
        v = np.array([1., 0])
        if plot: plt.plot(ucurrent[0], ucurrent[1], 'o')
        print("from", ucurrent, "in direction", v)
        left, right = ellipsoid_bracket(ucurrent, v, region.ellipsoid_center, region.ellipsoid_inv_axes, region.enlarge)
        uleft = ucurrent + v * left
        uright = ucurrent + v * right

        if plot: 
            plt.plot([uleft[0], uright[0]], [uleft[1], uright[1]], 'x-')
            
            plt.savefig('test_ellipsoid_bracket.pdf', bbox_inches='tight')
            plt.close()
        print("ellipsoid bracket:", left, right)
        assert left <= 0, left
        assert right >= 0, right
        
        assert_point_touches_ellipsoid(ucurrent, v, left, region.ellipsoid_center, region.ellipsoid_invcov, region.enlarge)
        assert_point_touches_ellipsoid(ucurrent, v, right, region.ellipsoid_center, region.ellipsoid_invcov, region.enlarge)

def test_crop_bracket(plot=False):
    ucurrent = np.array([0.39676747, 0.53881673])
    v = np.array([-0.79619985, -0.60503372])
    ellipsoid_center = np.array([0.23556461, 0.49899689])
    ellipsoid_inv_axes = np.array([[-3.28755896,  0.70136518], [ 1.33333377,  1.72933397]])
    enlarge = 26.66439694551674
    ellipsoid_invcov = np.array([[11.29995701, -3.17051875], [-3.17051875,  4.76837493]])
    #enlarge = 1.0
    #ellipsoid_inv_axes = np.array([[1.0, 0.], [0., 1]])
    

    eleft, eright = ellipsoid_bracket(ucurrent, v, ellipsoid_center, ellipsoid_inv_axes, enlarge)
    if plot:
        us = np.random.uniform(-2, +2, size=(10000, 2))
        d = us - ellipsoid_center
        r = np.einsum('ij,jk,ik->i', d, ellipsoid_invcov, d)
        mask_inside = r <= enlarge
        
        import matplotlib.pyplot as plt
        plt.plot(us[mask_inside,0], us[mask_inside,1], '+', ms=2)
        plt.plot(ucurrent[0], ucurrent[1], 'o ', ms=2)
        plt.plot([ucurrent[0] + eleft * v[0], ucurrent[0] + eright * v[0]],
            [ucurrent[1] + eleft * v[1], ucurrent[1] + eright * v[1]],
            '-s', ms=8)
        plt.savefig('test_crop_bracket.pdf', bbox_inches='tight')
    print("left:", eleft, ucurrent + v * eleft)
    assert eleft <= 0, eleft
    assert_point_touches_ellipsoid(ucurrent, v, eleft, ellipsoid_center, ellipsoid_invcov, enlarge)
    print("right:", eright, ucurrent + v * eright)
    assert eright >= 0, eright
    assert_point_touches_ellipsoid(ucurrent, v, eright, ellipsoid_center, ellipsoid_invcov, enlarge)

    left, right, cropleft, cropright = crop_bracket_at_unit_cube(ucurrent, v, eleft, eright)
    if plot:
        plt.plot([ucurrent[0] + left * v[0], ucurrent[0] + right * v[0]],
            [ucurrent[1] + left * v[1], ucurrent[1] + right * v[1]],
            's--', ms=8)
        plt.savefig('test_crop_bracket.pdf', bbox_inches='tight')
        plt.close()
    assert cropleft
    assert cropright
    assert (ucurrent + v * left <= 1).all(), (ucurrent, v, ellipsoid_center, ellipsoid_inv_axes, enlarge)
    assert (ucurrent + v * right <= 1).all(), (ucurrent, v, ellipsoid_center, ellipsoid_inv_axes, enlarge)
    assert (ucurrent + v * left >= 0).all(), (ucurrent, v, ellipsoid_center, ellipsoid_inv_axes, enlarge)
    assert (ucurrent + v * right >= 0).all(), (ucurrent, v, ellipsoid_center, ellipsoid_inv_axes, enlarge)

def test_aharm_sampler():
    def loglike(theta):
        return -0.5 * (((theta - 0.5)/0.01)**2).sum(axis=1)
    def transform(x):
        return x

    seed = 1
    Nlive = 10
    np.random.seed(seed)
    us = np.random.uniform(size=(Nlive, 2))
    Ls = loglike(us)
    Lmin = Ls.min()
    transformLayer = ScalingLayer()
    transformLayer.optimize(us, us)
    region = MLFriends(us, transformLayer)
    region.maxradiussq, region.enlarge = region.compute_enlargement()
    region.create_ellipsoid()
    assert region.inside(us).all()
    nsteps = 10
    sampler = AHARMSampler(nsteps=nsteps, region_filter=True)

    nfunccalls = 0
    ncalls = 0
    while True:
        u, p, L, nc = sampler.__next__(region, Lmin, us, Ls, transform, loglike)
        nfunccalls += 1
        ncalls += nc
        if u is not None:
            break
        if nfunccalls > 100 + nsteps:
            assert False, ('infinite loop?', seed, nsteps, Nlive)
    print("done in %d function calls, %d likelihood evals" % (nfunccalls, ncalls))
    

def run_aharm_sampler():
    for seed in [733] + list(range(10)):
        print()
        print("SEED=%d" % seed)
        print()
        np.random.seed(seed)
        nsteps = max(1, int(10**np.random.uniform(0, 3)))
        Nlive = int(10**np.random.uniform(1.5, 3))
        print("Nlive=%d nsteps=%d" % (Nlive, nsteps))
        sampler = AHARMSampler(nsteps, adaptive_nsteps=False, region_filter=False)
        us = np.random.uniform(0.6, 0.8, size=(4000, 2))
        Ls = loglike_vectorized(us)
        i = np.argsort(Ls)[-Nlive:]
        us = us[i,:]
        Ls = Ls[i]
        Lmin = Ls.min()
        
        transformLayer = ScalingLayer()
        transformLayer.optimize(us, us)
        region = MLFriends(us, transformLayer)
        region.maxradiussq, region.enlarge = region.compute_enlargement()
        region.create_ellipsoid()
        nfunccalls = 0
        ncalls = 0
        while True:
            u, p, L, nc = sampler.__next__(region, Lmin, us, Ls, transform, loglike)
            nfunccalls += 1
            ncalls += nc
            if u is not None:
                break
            if nfunccalls > 100 + nsteps:
                assert False, ('infinite loop?', seed, nsteps, Nlive)
        print("done in %d function calls, %d likelihood evals" % (nfunccalls, ncalls))


if __name__ == '__main__':
    #test_stepsampler_cubemh(plot=True)
    #test_stepsampler_regionmh(plot=False)
    #test_stepsampler_de(plot=False)
    #test_stepsampler_cubeslice(plot=True)
    #test_stepsampler_regionslice(plot=True)
    run_aharm_sampler()
    #test_ellipsoid_bracket()
    #test_crop_bracket(plot=True)



import numpy as np
import os
import matplotlib.pyplot as plt
from ultranest.mlfriends import ScalingLayer, AffineLayer, MLFriends
from numpy.testing import assert_allclose

here = os.path.dirname(__file__)


def test_region_sampling_scaling(plot=False):
    np.random.seed(1)
    upoints = np.random.uniform(0.2, 0.5, size=(1000, 2))
    upoints[:,1] *= 0.1
    
    transformLayer = ScalingLayer(wrapped_dims=[])
    transformLayer.optimize(upoints, upoints)
    region = MLFriends(upoints, transformLayer)
    region.maxradiussq, region.enlarge = region.compute_enlargement(nbootstraps=30)
    print("enlargement factor:", region.enlarge, 1 / region.enlarge)
    region.create_ellipsoid()
    nclusters = transformLayer.nclusters
    assert nclusters == 1
    assert np.allclose(region.unormed, region.transformLayer.transform(upoints)), "transform should be reproducible"
    assert region.inside(upoints).all(), "live points should lie near live points"
    if plot:
        plt.plot(upoints[:,0], upoints[:,1], 'x ')
        for method in region.sampling_methods:
            points, nc = method(nsamples=400)
            plt.plot(points[:,0], points[:,1], 'o ', label=str(method.__name__))
        plt.legend(loc='best')
        plt.savefig('test_regionsampling_scaling.pdf', bbox_inches='tight')
        plt.close()

    for method in region.sampling_methods:
        print("sampling_method:", method)
        newpoints = method(nsamples=4000)
        lo1, lo2 = newpoints.min(axis=0)
        hi1, hi2 = newpoints.max(axis=0)
        assert 0.15 < lo1 < 0.25, (method.__name__, newpoints, lo1, hi1, lo2, hi2)
        assert 0.015 < lo2 < 0.025, (method.__name__, newpoints, lo1, hi1, lo2, hi2)
        assert 0.45 < hi1 < 0.55, (method.__name__, newpoints, lo1, hi1, lo2, hi2)
        assert 0.045 < hi2 < 0.055, (method.__name__, newpoints, lo1, hi1, lo2, hi2)
        assert region.inside(newpoints).mean() > 0.99, region.inside(newpoints).mean()

    region.maxradiussq = 1e-90
    assert np.allclose(region.unormed, region.transformLayer.transform(upoints)), "transform should be reproducible"
    assert region.inside(upoints).all(), "live points should lie very near themselves"


def test_region_sampling_affine(plot=False):
    np.random.seed(1)
    upoints = np.random.uniform(size=(1000, 2))
    upoints[:,1] *= 0.5
    
    transformLayer = AffineLayer(wrapped_dims=[])
    transformLayer.optimize(upoints, upoints)
    region = MLFriends(upoints, transformLayer)
    region.maxradiussq, region.enlarge = region.compute_enlargement(nbootstraps=30)
    print("enlargement factor:", region.enlarge, 1 / region.enlarge)
    region.create_ellipsoid()
    nclusters = transformLayer.nclusters
    assert nclusters == 1
    assert np.allclose(region.unormed, region.transformLayer.transform(upoints)), "transform should be reproducible"
    assert region.inside(upoints).all(), "live points should lie near live points"
    if plot:
        plt.plot(upoints[:,0], upoints[:,1], 'x ')
        for method in region.sampling_methods:
            points, nc = method(nsamples=400)
            plt.plot(points[:,0], points[:,1], 'o ', label=str(method.__name__))
        plt.legend(loc='best')
        plt.savefig('test_regionsampling_affine.pdf', bbox_inches='tight')
        plt.close()

    for method in region.sampling_methods:
        print("sampling_method:", method)
        newpoints = method(nsamples=4000)
        lo1, lo2 = newpoints.min(axis=0)
        hi1, hi2 = newpoints.max(axis=0)
        assert 0 <= lo1 < 0.1, (method.__name__, newpoints, lo1, hi1, lo2, hi2)
        assert 0 <= lo2 < 0.1, (method.__name__, newpoints, lo1, hi1, lo2, hi2)
        assert 0.95 < hi1 <= 1, (method.__name__, newpoints, lo1, hi1, lo2, hi2)
        assert 0.45 <= hi2 < 0.55, (method.__name__, newpoints, lo1, hi1, lo2, hi2)
        assert region.inside(newpoints).all()

    region.maxradiussq = 1e-90
    assert np.allclose(region.unormed, region.transformLayer.transform(upoints)), "transform should be reproducible"
    assert region.inside(upoints).all(), "live points should lie very near themselves"

def test_region_ellipsoid(plot=False):
    np.random.seed(1)
    points = np.random.uniform(0.4, 0.6, size=(1000, 2))
    points[:,1] *= 0.5
    
    transformLayer = AffineLayer(wrapped_dims=[])
    transformLayer.optimize(points, points)
    region = MLFriends(points, transformLayer)
    region.maxradiussq, region.enlarge = region.compute_enlargement(nbootstraps=30)
    print("enlargement factor:", region.enlarge, 1 / region.enlarge)
    region.create_ellipsoid()
    nclusters = transformLayer.nclusters
    assert nclusters == 1
    
    bpts = np.random.uniform(size=(100, 2))
    mask = region.inside_ellipsoid(bpts)
    
    d = (bpts - region.ellipsoid_center)
    mask2 = np.einsum('ij,jk,ik->i', d, region.ellipsoid_invcov, d) <= region.enlarge
    
    assert_allclose(mask, mask2)


def test_region_mean_distances():
    np.random.seed(1)
    points = np.random.uniform(0.4, 0.6, size=(10000, 2))
    #points[:,1] *= 0.5
    mask = np.abs((points[:,0]-0.5)**2 + (points[:,1]-0.5)**2 - 0.08**2) < 0.02**2
    print('circle:', mask.sum())
    points = points[mask]
    mask = points[:,0] < 0.5
    print('half-circle:', mask.sum())
    points = points[mask]
    
    transformLayer = AffineLayer(wrapped_dims=[])
    transformLayer.optimize(points, points)
    region = MLFriends(points, transformLayer)
    region.maxradiussq, region.enlarge = region.compute_enlargement(nbootstraps=30)
    print("enlargement factor:", region.enlarge, 1 / region.enlarge)
    region.create_ellipsoid()
    meandist = region.compute_mean_pair_distance()
    
    t = transformLayer.transform(region.u)
    d = 0
    N = 0
    for i in range(len(t)):
        for j in range(i):
            d += ((t[i,:] - t[j,:])**2).sum()**0.5
            #print(i, j, t[i,:], t[j,:], ((t[i,:] - t[j,:])**2).sum())
            N += 1
    
    print((meandist, d, N, t))
    assert np.isclose(meandist, d / N), (meandist, d, N)


if __name__ == '__main__':
    test_region_sampling_scaling(plot=True)
    test_region_sampling_affine(plot=True)



import numpy as np

from ultranest.viz import round_parameterlimits

def wrap_single_test(vlo, vhi, plo_expected, phi_expected):
    assert vlo < vhi
    err_msg = 'for input values (%s, %s)' % (vlo, vhi)
    plo, phi, fmts = round_parameterlimits(np.asarray([vlo]), np.asarray([vhi]))
    np.testing.assert_allclose(plo, plo_expected, err_msg=err_msg)
    np.testing.assert_allclose(phi, phi_expected, err_msg=err_msg)

def wrap_single_fmt_test(vlo, vhi, fmt_expected):
    assert vlo < vhi
    plo, phi, fmts = round_parameterlimits(np.asarray([vlo]), np.asarray([vhi]), [(vlo, vhi)])
    assert fmts[0] == fmt_expected, (fmts, fmt_expected)
    fmt = fmts[0]
    assert fmt % plo != fmt % phi, (fmt, plo, phi, fmt % plo, fmt % phi)

def test_rounding_pos():
    wrap_single_test(0.00003, 0.001, 0, 0.001)
    wrap_single_test(0.1, 0.9, 0, 1)
    wrap_single_test(1.5, 150, 0, 1000)
    wrap_single_test(20000, 100000, 0, 100000)

def test_rounding_u():
    # test 0-1 range
    wrap_single_test(0, 1., 0, 1)
    wrap_single_test(0.0001, 0.99, 0, 1)
    wrap_single_test(0.001, 0.99, 0, 1)
    wrap_single_test(0.01, 0.9999, 0, 1)

def test_rounding_negpos():
    wrap_single_test(-0.1, 0.9, -1, 1)
    wrap_single_test(-1.5, 150, -1000, 1000)
    wrap_single_test(-20000, 100000, -100000, 100000)

def test_rounding_withguess():
    plo, phi, fmts = round_parameterlimits(
        np.asarray([-3.14, 0.01, 3000]), 
        np.asarray([0.9, 0.3, 100000]), 
        [(-3.14, 3.14), (0, 1.0), (-2000, 10000)])
    assert np.allclose(plo, [-3.14, 0, 0]), plo
    assert np.allclose(phi, [3.14, 1, 100000]), phi

    plo, phi, fmts = round_parameterlimits(
        np.asarray([1.4, 24.0, 0.4]),
        np.asarray([2.6, 25.5, 7.99]),
        [(1, 3), (20, 26), (0.1, 8.0)])
    assert np.allclose(plo, [1, 20, 0.1]), plo
    assert np.allclose(phi, [3, 26, 8]), phi

def test_fmt():
    wrap_single_fmt_test(-4.14, -4.13, "%+.3f")
    wrap_single_fmt_test(1.0, 3.0, "%+.1f")
    wrap_single_fmt_test(2000, 50000, "%+.1e")

if __name__ == '__main__':
    test_rounding_u()
    test_rounding_pos()
    test_rounding_negpos()
    test_rounding_withguess()
